{-# LANGUAGE DataKinds #-}
{-# LANGUAGE DeriveGeneric #-}
{-# LANGUAGE TypeOperators #-}
{-# LANGUAGE OverloadedStrings #-}
{-# LANGUAGE TemplateHaskell #-}
{-# LANGUAGE FlexibleInstances #-}
module ISIDORE.Client where


import Control.Monad
import Servant.API
import Servant.Client
import GHC.Generics (Generic)
import Data.Proxy(Proxy(..))
import Data.ByteString.Lazy (ByteString)
import Data.Aeson
import qualified Data.Vector as V
import Data.Text (Text)
import qualified Data.Text as T

data LangText =
  LangText { _lang :: Text
           , _text :: Text
           }
  | OnlyText Text
  | ArrayText [LangText]
  deriving (Show)

instance FromJSON LangText where
  parseJSON (Object o) = LangText <$> (o .: "@xml:lang") <*> (o .: "$")
  parseJSON (String x) = OnlyText <$> (pure x)
  parseJSON (Array  xs) = ArrayText <$> mapM parseJSON (V.toList xs)
  --parseJSON Null       = OnlyText <$> pure "null"


data Creator =
  Creator { _creator :: Author }
  | Creators { _creators :: [Author] }
  deriving (Show)

instance FromJSON Creator where
  parseJSON (Object o) = Creator <$> (o .: "creator")
  parseJSON (Array  os) = Creators <$> mapM parseJSON (V.toList os)

data Author =
  Author { _firstName :: Name
         , _lastName  :: Name
         }
  | Authors [Author]
  deriving (Show)

data Name = Name { _name :: Text }
  deriving (Show)

instance FromJSON Name where
  parseJSON (String o) = Name <$> (pure o)
  parseJSON (Array  os) = Name <$> (pure "empty")


instance FromJSON Author where
  parseJSON (Object  o) = Author <$> (o .: "firstname") <*> (o .: "lastname")
  parseJSON (Array  os) = Authors <$> mapM parseJSON (V.toList os)

data SourceInfo =
  SourceInfo { _shortLabel :: Text
             , _handle     :: Text
             , _collectionUuid :: Text
             , _sourceName :: Text }
    deriving (Show)


instance FromJSON SourceInfo where
  parseJSON (Object o) =
    SourceInfo <$> (o .: "@shortLabel")
               <*> (o .: "@handle")
               <*> (o .: "@collectionUuid")
               <*> (o .: "$")


data IsidoreDoc = IsidoreDoc
  { _title    :: LangText
  , _abstract :: Maybe LangText
  , _date     :: Maybe Text
  , _url      :: Maybe Text
  , _source   :: Maybe SourceInfo
  , _authors  :: Maybe Creator
  }
  deriving (Show)


data Reply = Reply { _docs :: [IsidoreDoc]}
  deriving (Show)

instance FromJSON IsidoreDoc where
  parseJSON (Object o) = IsidoreDoc <$> (i >>= (.: "title"       ))
                                    <*> (i >>= (.:? "abstract"   ))
                                    <*> (i >>= (.: "date"        )     >>= (.:? "normalizedDate" ))
                                    <*> (i >>= (.:? "url"        ))
                                    <*> (i >>= (.: "source_info" )     >>= (.:? "collectionLabel"))
                                    <*> (i >>= (.:? "enrichedCreators"))
    where
      i = (o .: "isidore")



instance FromJSON Reply where
  parseJSON (Object o) = Reply <$> (responseReplies >>= parseJSON)
                         -- <*> (responseReplies >>= (.: "abstract") >>= (.: "$"))
                         -- <*> (responseReplies >>= (.: "date") >>= (.: "origin"))
                         -- <*> (responseReplies >>= (.: "enrichedCreators") >>= (.: "creator"))
    where
      responseReplies = (o .: "response")
            >>= (.: "replies")
            >>= (.: "content")
            >>= (.: "reply")
      -- docs = mapM (\r -> IsidoreDoc <$> (r .: "isidore") >>= (.: "title")) responseReplies
          -- >>= (.: "$"))
          -- >>= (.: "isidore")

data Output = JSON

instance ToHttpApiData Output where
  toUrlPiece JSON = "json"

type ISIDOREAPI = Search

-- search?q=colza&output=json&replies=1&author=jm&
type Search = "search"
  :> QueryParam "replies" Int
  :> QueryParam "output" Output
  :> QueryParam "q" T.Text
  :> QueryParam "author" T.Text
  :> Get '[JSON] Reply

isidoreAPI :: Proxy ISIDOREAPI
isidoreAPI = Proxy

search :: Maybe Int -> Maybe T.Text -> Maybe T.Text -> ClientM Reply
search n q a = client isidoreAPI n (Just JSON) q a


